#ifndef ATOM_PARSER_H_INCLUDED
#define ATOM_PARSER_H_INCLUDED

#include <expat.h>
#include <stdio.h>

#include <iterator>
#include <map>
#include <string>
#include <list>
#include <cassert>

/**
   Parser state information.
 */
class Atom_parser {
public:
  typedef std::basic_string<XML_Char> Tag;
  typedef std::basic_string<XML_Char> Text;

  struct Entry {
    void clear() {
      title.clear();
      link.clear();
      updated.clear();
      id.clear();
      author.clear();
      summary.clear();
    }

    /** DDL for defining the table */
    static const char* const ddl;

    std::string title;
    std::string link;
    std::string updated;
    std::string id;
    std::string summary;
    std::string author;
  };

  struct Feed {
    std::string title;
    std::string subtitle;
    std::string link;
    std::string updated;
    std::string author;
  };

  enum Error_code {
    ATOM_ERR_NONE,
    ATOM_ERR_NOMEM,
    ATOM_ERR_PARSE,
    ATOM_ERR_FILE_READ,
    ERROR_CODE_COUNT
  };

public:
  Atom_parser(FILE* file);

  ~Atom_parser() {
    XML_ParserFree(m_parser);
  }

  int fill_from_file();

  bool eof()   const { return m_eof; }
  bool empty() const { return m_entries.empty() && m_eof; }

  void advance_entry() {
    if (!m_entries.empty())
      m_entries.pop_front();
    fill_from_file();
  }

  const Entry* current_entry() const {
    assert(!m_entries.empty());
    return &m_entries.front();
  }

  Entry* current_entry() {
    assert(!m_entries.empty());
    return &m_entries.front();
  }

private:
  enum Tag_type { TAG_TYPE_START, TAG_TYPE_END };

  typedef void (Atom_parser::*Handler)(const XML_Char **, Tag_type);
  typedef std::map<Tag, Handler> Action_map;

  static const Action_map::value_type s_action_list[];
  static const Action_map s_action;

  static void start_handler(void*, const XML_Char*, const XML_Char**);
  static void end_handler(void*, const XML_Char*);
  static void text_handler(void*, const XML_Char*, int);

  Handler find_handler(const XML_Char *name) const;

  void author_tag(const XML_Char **atts, Tag_type tag_type);
  void entry_tag(const XML_Char **atts, Tag_type tag_type);
  void feed_tag(const XML_Char **atts, Tag_type tag_type);
  void id_tag(const XML_Char **atts, Tag_type tag_type);
  void link_tag(const XML_Char **atts, Tag_type tag_type);
  void summary_tag(const XML_Char **atts, Tag_type tag_type);
  void title_tag(const XML_Char **atts, Tag_type tag_type);
  void updated_tag(const XML_Char **atts, Tag_type tag_type);

  std::list<Entry> m_entries;      /**< List for holding entries */

  XML_Parser m_parser;
  FILE* m_file;

  Feed m_feed;                     /**< Feed data */
  bool m_eof;                      /**< End-of-file marker */
  Entry m_pending;                 /**< Unfilled entry */
  bool m_is_in_entry;              /**< Marker if we are inside an entry or not */
  std::string* m_assign_to;        /**< Where text data should be stored when encountered */
};

#endif /* ATOM_PARSER_H_INCLUDED */
